; patch.lisp
; (load "patch.lisp")
; Bill's bug

(in-package "COMPILER")

(push '((fixnum) boolean #.(flags)
  "1")
  (get 'integerp 'inline-always))

(in-package "SI")

(defun match-dimensions (dim pat)
  (if (null dim)
      (null pat)
      (and (or (eq (car pat) '*)
	       (eql (car dim) (car pat)))
	   (match-dimensions (cdr dim) (cdr pat)))))

(compile 'si::match-dimensions)

(in-package "ACL2")

(defun make-defstobj-declare-form (form)
  (let* ((name (cadr form))
         (args (cddr form))
         (template (defstobj-template name args))
         (raw-defs (defstobj-raw-defs name template)))
    (cons 'progn
          (mapcar (function
                   (lambda (def) (make-defun-declare-form (cons 'defun def))))
                  raw-defs))))

(defun proclaim-form (form)
  (cond ((consp form)
         (case (car form)
               ((in-package) (eval form) nil)
               ((defmacro defvar defparameter) nil)
               ((defconst)
                (eval (make-defconst-declare-form form))
                nil)
               ((defstobj)
                (eval (make-defstobj-declare-form form)))
               ((eval-when)
                (dolist (x (cddr form))
                        (proclaim-form x))
                nil)
               ((progn mutual-recursion)
                (dolist (x (cdr form))
                        (proclaim-form x))
                nil)
               ((defun)
                (eval (make-defun-declare-form form))
                nil)
               (otherwise nil)))
        (t nil)))

(defmacro defstobj (name &rest args)

; This function is run when we evaluate (defstobj name . args) in raw lisp.
; A typical such form is

  #|
  (defstobj $st
    (flag :type t :initially run)
    (pc   :type (integer 0 255) :initially 128)
    (mem  :type (array (integer 0 255) (256)) :initially 0))
  |#

; This function must contend with a problem analogous to the one
; addressed by acl2::defconst in acl2.lisp: the need to avoid
; re-declaration of the same stobj.  We use
; redundant-raw-lisp-discriminator in much the same way as in the raw
; lisp defmacro of acl2::defconst.

  (let* ((template (defstobj-template name args))
         (init (defstobj-raw-init name template))
         (the-live-name (the-live-var name)))
    `(progn
       (defvar ,the-live-name)

       ;;; Now we lay down the defuns of the recognizers, accessors
       ;;; and updaters as generated by defstob-raw-defs.  The
       ;;; boilerplate below just adds the DEFUN to the front of each
       ;;; def generated, preserving the order of the defs as
       ;;; generated.

       ,@(mapcar (function (lambda (def) (cons 'DEFUN def)))
                 (defstobj-raw-defs name template))

       (let ((template ',template))
         (cond
          ((boundp ',the-live-name)
           (let ((d (get ',the-live-name 'redundant-raw-lisp-discriminator)))

; d is expected to be of the form (DEFSTOBJ namep . field-templates)

             (cond
              ((and (consp d)
                    (eq (car d) 'defstobj)
                    (consp (cdr d))
                    (eq (cadr d) (car template))
                    (equal (cddr d) (cadr template))
                    (stobj-initial-statep (symbol-value ',the-live-name)
                                          (cadr template)))
               ',name)
              (t (error
                  "Illegal attempt to redeclare the single-threaded ~
               object ~s."
                  ',name)))))
          (t
           (defparameter ,the-live-name ,init)
           (setf (get ',the-live-name 'redundant-raw-lisp-discriminator)
                 (list* 'defstobj (car template) (cadr template)))
           (setq *user-stobj-alist*
                 (cons (cons ',name (symbol-value ',the-live-name))
                       *user-stobj-alist*))
           (setq ,name ,the-live-name)
           ',name))))))

(compile
 (defun immediate-forcep (fn ens)
  (cond ((eq fn 'case-split) 'case-split)
        ((enabled-numep *immediate-force-modep-xnume* ens) t)
        (t nil))))

