; Following are examples of how we "wrap" functions calls in type
; declarations to guarantee that GCL understands that it can optimize
; them at compile time.

(defmacro +<32> (a b)
  `(Int32 (+ (Int32 ,a) (Int32 ,b))))

(defmacro lognot<32> (a)
  `(Int32 (lognot (Int32 ,a))))

(defmacro logand<32> (a b)
  `(Int32 (logand (Int32 ,a) (Int32 ,b))))

(defmacro logior<32> (a b)
  `(Int32 (logior (Int32 ,a) (Int32 ,b))))

(defun fix<5> (x)
  (declare (xargs :guard (integerp x)))
  (logext 5 x))

(defstobj st
  (CNTL :type (signed-byte 6) :initially 3)
  (DATA :type (signed-byte 6) :initially 4)
  )

; Here is a simple "reader" that provides a slightly more convenient
; means of programming with single threaded objects.

(defmacro progn-st (&rest args)

  (if (consp args)

      (let ((form (car args)))

	(case (first form)

	      ((Int32)
	       `(let ((,(second form) ,(third form)))
		  (declare (type (signed-byte 32) ,(second form)))
		  (progn-st ,@(cdr args))))

	      (t
	       `(let ((st ,form))
		  (progn-st ,@(cdr args))))))

    `st))

(defthm stp-set-nth
  (implies
   (stp st)
   (and
    (equal (stp (set-nth 0 v st)) (signed-byte-p 6 v))
    (equal (stp (set-nth 1 v st)) (signed-byte-p 6 v))))
  :hints (("goal" :in-theory (enable signed-byte-p))))



(defun step-st (st)

  (declare (xargs :guard-hints (("goal" :in-theory (disable set-nth nth the-error)))
		  :stobjs (st)
		  :guard (stp st)))

  (progn-st

   (Int32 key (fix<5> (CNTL st)))

   (Int32 data (fix<5> (DATA st)))

   (case (logand<32> key #x0F)

	 ((#x00) (setDATA (+<32> data #x00) st))
	 ((#x01) (setDATA (+<32> data #x01) st))
	 ((#x02) (setDATA (+<32> data #x02) st))
	 ((#x03) (setDATA (+<32> data #x03) st))
	 ((#x04) (setDATA (+<32> data #x04) st))
	 ((#x05) (setDATA (+<32> data #x05) st))
	 ((#x06) (setDATA (+<32> data #x06) st))
	 ((#x07) (setDATA (+<32> data #x07) st))
	 ((#x08) (setDATA (+<32> data #x08) st))
	 ((#x09) (setDATA (+<32> data #x09) st))
	 ((#x0A) (setDATA (+<32> data #x0A) st))
	 ((#x0B) (setDATA (+<32> data #x0B) st))
	 ((#x0C) (setDATA (+<32> data #x0C) st))
	 ((#x0D) (setDATA (+<32> data #x0D) st))
	 ((#x0E) (setDATA (+<32> data #x0E) st))
	 ((#x0F) (setDATA (+<32> data #x0F) st))
	 (t         st)

	 )

   (if (< key 0)

       (setDATA (+<32> (DATA st) (lognot<32> #x0F)) st)
     
     st)

   (if (< key 0)

       (setCNTL (+<32> key 1) st)
     
     (setCNTL (+<32> key -1) st))

   )
  )


(defun alt-step-st (st)
  (declare (xargs :stobjs (st)
		  :guard (stp st)
		  :guard-hints (("goal" :in-theory (disable set-nth nth the-error)))))
  (progn-st

   (setDATA (+<32> (fix<5> (DATA st)) (fix<5> (CNTL st))) st)

   (Int32 key (fix<5> (CNTL st)))

   (if (< key 0)

       (setCNTL (+<32> key 1) st)
     
     (setCNTL (+<32> key -1) st))

   )
  ))

